package apollo.ares.media;
/**
 * @author Konrad Rzeszutek (konrad@darnok.org)
 * @version $Id: RTPProcessor.java,v 1.2 2002/04/28 19:14:15 uid500 Exp $
 */

import apollo.common.*;
import apollo.common.tui.*;
import apollo.common.events.*;
import java.awt.Component;
import java.io.*;
import java.awt.*;
import javax.media.*;
import javax.media.format.*;
import javax.media.protocol.*;
import javax.media.protocol.DataSource;
import javax.media.datasink.*;
import javax.media.control.TrackControl;

import javax.media.control.QualityControl;
/**
 * RTP engine.
 */
public class RTPProcessor implements ControllerListener {

    private Processor p;
    //private DataSource ds;
    private Object waitSync = new Object();
    private boolean stateTransitionOK = true;
    private Log log = null;


    /**
     * Intializes this object
     *
     * @param log Logging functionality.
     *
     */
    public RTPProcessor(Log log) {

      this.log = log;

    }

    /**
     * Opens the stream from the specified by the <code>ds</code> DataSource.
     *
     * @param ds != null
     * @throws MalformedURLException if the <code>inputURL</code> is invalid.
     * @throws NoProcessorException is thrown when Manager can't find a Processor for a particular URL, MediaLocator or ProcessorModel
     * @throws NoDataSourceException is thrown when a DataSource can't be found for a particular URL or MediaLocator
     * @throws UnknownProblemException if something terribly wrong happend
     * @throws IOException if there are problems reading the data
     * @throws IncompatibleSourceException if the input format is a wrong one and not the one that's expected.
     * @throws UnsupportedFormatException if the format of the input media does not match the one that's required.
     */
    public void open(DataSource ds) throws NoProcessorException,
      UnknownProblemException, IOException, UnsupportedPlugInException,
      IncompatibleSourceException, UnsupportedFormatException {


        p = Manager.createProcessor(ds);
	p.addControllerListener(this);

	// Put the Processor into configured state.
	p.configure();
	if (!waitForState(p.Configured)) {
            throw new UnknownProblemException("Failed to configure the processor.");
	}
        // Add the effect codec.
        TrackControl tracks[] = p.getTrackControls();
        boolean programmed = false;

        // Search through the tracks for a video track
        for (int i = 0; i < tracks.length; i++) {
            Format format = tracks[i].getFormat();
            if (  tracks[i].isEnabled() &&
                  format instanceof VideoFormat &&
                  !programmed) {

                // Found a video track. Try to program it to output JPEG/RTP
                // Make sure the sizes are multiple of 8's.
                Dimension size = ((VideoFormat)format).getSize();
                float frameRate = ((VideoFormat)format).getFrameRate();
                int w = (size.width % 8 == 0 ? size.width :
                                (int)(size.width / 8) * 8);
                int h = (size.height % 8 == 0 ? size.height :
                                (int)(size.height / 8) * 8);
                VideoFormat jpegFormat = new VideoFormat(VideoFormat.JPEG_RTP,
                                                         new Dimension(w, h),
                                                         Format.NOT_SPECIFIED,
                                                         Format.byteArray,
                                                         frameRate);
                tracks[i].setFormat(jpegFormat);
                System.err.println("Video transmitted as:");
                System.err.println("  " + jpegFormat);
                // Assume succesful
                programmed = true;
            } else
                tracks[i].setEnabled(false);
        }

        if (!programmed)
          throw new UnknownProblemException("Input datasource does not contain"+
           " any video tracks.");

        // Set the output content descriptor to RAW_RTP

        p.setContentDescriptor(new ContentDescriptor(ContentDescriptor.RAW_RTP));
	p.realize();

        // Set the JPEG quality to .5.
        setJPEGQuality(p, 0.5f);

	if (!waitForState(p.Realized)) {
          throw new UnknownProblemException("Failed to realize the processor.");
	}
	p.start();
    }

        /**
     * Setting the encoding quality to the specified value on the JPEG encoder.
     * 0.5 is a good default.
     */
    private void setJPEGQuality(Player p, float val) {

	Control cs[] = p.getControls();
	QualityControl qc = null;
	VideoFormat jpegFmt = new VideoFormat(VideoFormat.JPEG);

	// Loop through the controls to find the Quality control for
 	// the JPEG encoder.
	for (int i = 0; i < cs.length; i++) {

	    if (cs[i] instanceof QualityControl &&
		cs[i] instanceof Owned) {
		Object owner = ((Owned)cs[i]).getOwner();

		// Check to see if the owner is a Codec.
		// Then check for the output format.
		if (owner instanceof Codec) {
		    Format fmts[] = ((Codec)owner).
                      getSupportedOutputFormats(null);
		    for (int j = 0; j < fmts.length; j++) {
			if (fmts[j].matches(jpegFmt)) {
			    qc = (QualityControl)cs[i];
	    		    qc.setQuality(val);
			    System.err.println("- Setting quality to " +
					val + " on " + qc);
			    break;
			}
		    }
		}
		if (qc != null)
		    break;
	    }
	}
    }

    /**
     * Stops the engine.
     */
    public void stop() {

      if (p != null)
          p.stop();
    }
    /**
     * Closes down the engine.
     */
    public void close() {

      if (p != null)
        p.close();

      p = null;
      System.gc();
    }

    /**
     * Get the visual component from the engine.
     */
    public Component getVisualComponenet() {
      if (p != null)
        return p.getVisualComponent();
      return null;
    }
    /**
     * Get the control component from the engine.
     *
     */
    public Component getControlPanelComponent() {

      if (p != null)
        return p.getControlPanelComponent();
      return null;
    }
    public DataSource getDataOutput() {

      if (p != null)
        return p.getDataOutput();
      return null;
    }

    /**
     * Block until the processor has transitioned to the given state.
     * Return false if the transition failed.
     */
    private boolean waitForState(int state) {
	synchronized (waitSync) {
	    try {
		while (p.getState() != state && stateTransitionOK)
		    waitSync.wait();
	    } catch (Exception e) {}
	}
	return stateTransitionOK;
      }

    /**
     * Controller Listener.
     */
    public void controllerUpdate(ControllerEvent evt) {

        System.out.println(this.getClass().getName()+" "+evt);
	if (evt instanceof ConfigureCompleteEvent ||
	    evt instanceof RealizeCompleteEvent ||
	    evt instanceof PrefetchCompleteEvent) {
	    synchronized (waitSync) {
		stateTransitionOK = true;
		waitSync.notifyAll();
	    }
	} else if (evt instanceof ResourceUnavailableEvent) {
	    synchronized (waitSync) {
		stateTransitionOK = false;
		waitSync.notifyAll();
	    }
	} else if (evt instanceof EndOfMediaEvent) {
	    p.close();
	    //System.exit(0);
	}
    }
}