package apollo.common.movie;

/**
 * @author Konrad Rzeszutek (konrad@darnok.org)
 * @version $Id: JPGtoMOV.java,v 1.4 2002/04/28 19:14:21 uid500 Exp $
 */

import java.io.*;
import java.util.*;
import java.awt.Dimension;

import javax.media.*;
import javax.media.control.*;
import javax.media.protocol.*;
import javax.media.protocol.DataSource;
import javax.media.datasink.*;
import javax.media.format.*;
import apollo.common.*;
import java.net.*;
import apollo.common.events.*;
/**
 * Converts a list of JPG files into a QuickTime (.MOV) movie file.
 *
 *
 */
public class JPGtoMOV implements   ControllerListener, DataSinkListener {


  /**
   * Frames per second. By default its 30 fps.
   */
  public int frameRate = 30;
  /**
   * The height of the movie and the images.
   */
  public int height    = 0;
  /**
   * The width of the movie and the images.
   */
  public int width     = 0;
  protected Log log = null;
  protected String LOG = null;
  private Object waitSync = new Object();
  private boolean stateTransitionOK = true;

  private MediaLocator out;

  /**
   * JPG to MOV (QuickTime) convertor.
   *
   * @param log The logging facility
   * @param output The absolute path to the movie file (including the extension)
   * @param width The width of the movie.
   * @param height  The height of the movie.
   *
   * @throws MalformedURLException if the given <code>output</code> is malformed (uses invalid characters)
   */
  public JPGtoMOV(Log log, String output, int width, int height) throws
  MalformedURLException {

    this.log = log;
    this.width = width;
    this.height = height;
    if ((out = createMediaLocator(output)) == null)
      throw new MalformedURLException( "Cannot build media locator from: "
      + output);
    LOG  = "MovieCreator";

  }

    /**
     * Create a media locator from the given URL
     *
     * @param url The URL to be used
     * @return the URL in a MediaLocator format.
     */
  private static MediaLocator createMediaLocator(String url) {

	MediaLocator ml;

	if (url.indexOf(":") > 0 && (ml = new MediaLocator(url)) != null)
	    return ml;

	if (url.startsWith(File.separator)) {
	    if ((ml = new MediaLocator("file:" + url)) != null)
		return ml;
	} else {
	    String file = "file:" + System.getProperty("user.dir") +
             File.separator + url;
	    if ((ml = new MediaLocator(file)) != null)
		return ml;
	}

	return null;
    }

    /**
     * Start converting the list of given <code>inFile</code> into the
     * an movie file (which was defined in the constructor).
     *
     * @param inFiles the list of absolute paths (and names) to the images. (The vector should be
     * a list of Strings).
     *
     * @throws IOException if the files have problems being read
     * @throws UnsupportedFormatException if the input files are not the JPG format
     * @throws UnknownProblemException if there the engine cannot be realized.
     * @throws NoProcessorException is thrown when Manager can't find a Processor for a particular URL, MediaLocator or ProcessorModel.
     * @throws MediaException indicates an unexpected error condition in a JMF method.
     */
    public void doIt(Vector inFiles) throws IOException,
          UnsupportedFormatException, UnknownProblemException,
          NoProcessorException, MediaException {

	ImageDataSource ids = new ImageDataSource(inFiles, this);
	Processor p;

        //log.println("Creating processor ..", LOG);
        p = Manager.createProcessor(ids);

	p.addControllerListener(this);

	// Put the Processor into configured state so we can set
	// some processing options on the processor.
	p.configure();
	if (!waitForState(p, p.Configured)) {
            throw new UnknownProblemException("Failed to configure the processor.");
	    //return false;
	}

	// Set the output content descriptor to QuickTime.
	p.setContentDescriptor(new ContentDescriptor(FileTypeDescriptor.QUICKTIME));

	// Query for the processor for supported formats.
	// Then set it on the processor.
	TrackControl tcs[] = p.getTrackControls();
	Format f[] = tcs[0].getSupportedFormats();
	if (f == null || f.length <= 0)
            throw new UnsupportedFormatException("The mux does not support the"+
              " input format", tcs[0].getFormat());

	tcs[0].setFormat(f[0]);

	//log.println("Setting the track format to: " + f[0],LOG);

	// We are done with programming the processor.  Let's just
	// realize it.
	p.realize();
	if (!waitForState(p, p.Realized))
            throw new UnknownProblemException("Failed to realize the processor.");

	// Now, we'll need to create a DataSink.
	DataSink dsink;
	if ((dsink = createDataSink(p, out)) == null)
            throw new MediaException("Failed to create a DataSink for the" +
            " given output: "+out.toString());

	dsink.addDataSinkListener(this);
	fileDone = false;
        log.println(" - Making a movie out of  "+inFiles.size()+" files.",LOG);
	//log.println("start processing...",LOG);

        p.start();
        dsink.start();

	waitForFileDone();

        dsink.close();

	p.removeControllerListener(this);

	log.println("Done.",LOG);

    }


    /**
     * Create the DataSink.
     */
    private DataSink createDataSink(Processor p, MediaLocator outML)
      throws NoDataSinkException, IOException {

	DataSource ds;

	if ((ds = p.getDataOutput()) == null)

            throw new javax.media.NoDataSinkException("Something is really"+
            " wrong: the processor does not have an output DataSource: "+outML);

	DataSink dsink;

        Object context = null;
        log.println("- Writting to: " + outML,LOG);
        dsink = Manager.createDataSink(ds, outML);
        dsink.open();

	return dsink;
    }



    /**
     * Block until the processor has transitioned to the given state.
     * Return false if the transition failed.
     */
    private boolean waitForState(Processor p, int state) {
	synchronized (waitSync) {
	    try {
		while (p.getState() < state && stateTransitionOK)
		    waitSync.wait();
	    } catch (Exception e) {}
	}
	return stateTransitionOK;
    }


    /**
     * Controller Listener.
     */
    public void controllerUpdate(ControllerEvent evt) {

	if (evt instanceof ConfigureCompleteEvent ||
	    evt instanceof RealizeCompleteEvent ||
	    evt instanceof PrefetchCompleteEvent) {
	    synchronized (waitSync) {
		stateTransitionOK = true;
		waitSync.notifyAll();
	    }
	} else if (evt instanceof ResourceUnavailableEvent) {
	    synchronized (waitSync) {
		stateTransitionOK = false;
		waitSync.notifyAll();
	    }
	} else if (evt instanceof EndOfMediaEvent) {
	    evt.getSourceController().stop();
	    evt.getSourceController().close();
	}
    }


    private Object waitFileSync = new Object();
    private boolean fileDone = false;
    private boolean fileSuccess = true;

    /**
     * Block until file writing is done.
     */
    private boolean waitForFileDone() {
	synchronized (waitFileSync) {
	    try {
		while (!fileDone)
		    waitFileSync.wait();
	    } catch (Exception e) {}
	}
	return fileSuccess;
    }


    /**
     * Event handler for the file writer.
     */
    public void dataSinkUpdate(DataSinkEvent evt) {

	if (evt instanceof EndOfStreamEvent) {
	    synchronized (waitFileSync) {
		fileDone = true;
		waitFileSync.notifyAll();
	    }
	} else if (evt instanceof DataSinkErrorEvent) {
	    synchronized (waitFileSync) {
		fileDone = true;
		fileSuccess = false;
		waitFileSync.notifyAll();
	    }
	}
    }


  	// Generate the output media locators.



    ///////////////////////////////////////////////
    //
    // Inner classes.
    ///////////////////////////////////////////////


    /**
     * A DataSource to read from a list of JPEG image files and
     * turn that into a stream of JMF buffers.
     * The DataSource is not seekable or positionable.
     */
    class ImageDataSource extends PullBufferDataSource {

	ImageSourceStream streams[];

	ImageDataSource(Vector images, JPGtoMOV parent) {
	    streams = new ImageSourceStream[1];
	    streams[0] = new ImageSourceStream(images, parent);
	}

	public void setLocator(MediaLocator source) {
	}

	public MediaLocator getLocator() {
	    return null;
	}

	/**
	 * Content type is of RAW since we are sending buffers of video
	 * frames without a container format.
	 */
	public String getContentType() {
	    return ContentDescriptor.RAW;
	}

	public void connect() {
	}

	public void disconnect() {
	}

	public void start() {
	}

	public void stop() {
	}

	/**
	 * Return the ImageSourceStreams.
	 */
	public PullBufferStream[] getStreams() {
	    return streams;
	}

	/**
	 * We could have derived the duration from the number of
	 * frames and frame rate.  But for the purpose of this program,
	 * it's not necessary.
	 */
	public Time getDuration() {
	    return DURATION_UNKNOWN;
	}

	public Object[] getControls() {
	    return new Object[0];
	}

	public Object getControl(String type) {
	    return null;
	}
    }


    /**
     * The source stream to go along with ImageDataSource.
     */
    class ImageSourceStream implements PullBufferStream {

	Vector images;
        private JPGtoMOV p;

	VideoFormat format;

	int nextImage = 0;	// index of the next image to be read.
	boolean ended = false;

	public ImageSourceStream(Vector images, JPGtoMOV parent) {
	    this.images = images;
            this.p = parent;
	    format = new VideoFormat(VideoFormat.JPEG,
				new Dimension(p.width, p.height),
				Format.NOT_SPECIFIED,
				Format.byteArray,
				(float)p.frameRate);

	}

	/**
	 * We should never need to block assuming data are read from files.
	 */
	public boolean willReadBlock() {
	    return false;
	}

	/**
	 * This is called from the Processor to read a frame worth
	 * of video data.
	 */
 	public void read(Buffer buf) throws IOException {

	    // Check if we've finished all the frames.
	    if (nextImage >= images.size()) {
		// We are done.  Set EndOfMedia.
		p.log.println(" - Done reading all images.",p.LOG);
		buf.setEOM(true);
		buf.setOffset(0);
		buf.setLength(0);
		ended = true;
		return;
	    }

	    String imageFile = (String)images.elementAt(nextImage);
	    nextImage++;

	    // Open a random access file for the next image.
	    RandomAccessFile raFile;
	    raFile = new RandomAccessFile(imageFile, "r");

	    byte data[] = null;

	    // Check the input buffer type & size.

	    if (buf.getData() instanceof byte[])
		data = (byte[])buf.getData();

	    // Check to see the given buffer is big enough for the frame.
	    if (data == null || data.length < raFile.length()) {
		data = new byte[(int)raFile.length()];
		buf.setData(data);
	    }

	    // Read the entire JPEG image from the file.
	    raFile.readFully(data, 0, (int)raFile.length());

	    buf.setOffset(0);
	    buf.setLength((int)raFile.length());
	    buf.setFormat(format);
	    buf.setFlags(buf.getFlags() | buf.FLAG_KEY_FRAME);

	    // Close the random access file.
	    raFile.close();
	}

	/**
	 * Return the format of each video frame.  That will be JPEG.
	 */
	public Format getFormat() {
	    return format;
	}

	public ContentDescriptor getContentDescriptor() {
	    return new ContentDescriptor(ContentDescriptor.RAW);
	}

	public long getContentLength() {
	    return 0;
	}

	public boolean endOfStream() {
	    return ended;
	}

	public Object[] getControls() {
	    return new Object[0];
	}

	public Object getControl(String type) {
	    return null;
	}
    }
}