package apollo.common.movie;

/**
 * @author Konrad Rzeszutek
 * @version $Id: MovieCreator.java,v 1.6 2002/04/28 19:14:21 uid500 Exp $
 */
import apollo.common.*;
import apollo.common.client.*;
import java.io.*;
import com.sun.image.codec.jpeg.*;
import java.awt.image.*;
import java.util.*;
import apollo.demeter.client.*;
import java.text.*;
import apollo.common.events.*;
import javax.media.MediaLocator;
/**
 * Movie generator. Assembles movies (using {@Link JPGtoMOV})
 */
public class MovieCreator extends Thread {

  /**
   * How often the movies should be generated. The number is in hours.
   * The default is every 2 hours.
   */
  public int MOVIE_CREATION_TIME = 1; // Every two hours

  /**
   * How often the thread should check to see if its time to create a movie.
   * Default its 600000 miliseconds (or 10 minutes).
   *
   */
  public int SLEEP_TIME = 10*6000; // Every 10 minutes.
  public Log log;
  private boolean stop = false;
  private Vector clients = new Vector();
  private Object waitThread = new Object();

  private SimpleDateFormat sdf = new java.text.SimpleDateFormat("MMddyy-HHmm");

  /**
   * Initialize the logging object.
   *
   */
  public MovieCreator(Log log) {
    this.log = log;
    this.start();
  }

  /**
   * Starts processing (assembling a movie) on the given <code>c</code> client.
   *
   * @param c The client on which the movie must be
   * @throws IsNotRegisteredException if the client has not registered using the
   *  register(DemeterClient) method
   */
  public synchronized void make(DemeterClient c) throws IsNotRegisteredException{

    //log.println("Start assembling a movie for "+c.getHost()+".",c.getHost());

    int i = clients.indexOf(c);
    if (i < 0) throw new IsNotRegisteredException("Client is not registered!");

    String date = sdf.format(new Date());

    log.println("Starting creating a movie in "+c.getMovieDir()+
      " with name: "+date+".mov",Log.INFO);

    try {
      new Worker(this, c.getImageDir(), c.getMovieDir()+"/"+date+".mov");
    } catch (Exception e) {
      log.normalError(e.getMessage());
    }

  }

  /**
   * Stops processing and removes the given client from the movie-making
   * assembly line.
   *
   * @param  c The client to remove.
   */
  public  synchronized void remove(DemeterClient c) {
      log.println("- Removing "+c.getHost()+" from the movie-generator.",c.getHost());

      clients.remove(c);
  }

  /**
   * Flush and close all the clients.
   */
  public void closeAll() {

    Iterator i = clients.iterator();
    while (i.hasNext()) {
      DemeterClient c = (DemeterClient)i.next();
      flush(c);

    }
     i = clients.iterator();
    while (i.hasNext()) {
      DemeterClient c = (DemeterClient)i.next();
      remove(c);
    }

  }
  /**
   * Flush (or speed up) the movie-generation. This is usually done right
   * before removing the client.
   *
   * @param c The client to flush.
   */
  public void flush(DemeterClient c) {

      log.println("- Flushing "+c.getHost()+".",c.getHost());
      try {
        make(c);
      } catch (Exception e) { }
  }

  /**
   * Register a new client.
   *
   * @param c The new client.
   */
  public synchronized void register(DemeterClient c) {

      log.println("- Registered a new client "+c.getHost()+".",c.getHost());
      clients.add(c);
      c.setTimer(0);
  }

  /**
   * Terminate the thread safely.
   *
   */
  public void killThread() {

    stop = true;
    synchronized (waitThread) {
      try {
        waitThread.notifyAll();
      } catch (Exception e) { }
    }

  }
  public void run() {


    DemeterClient c = null;

    while (!stop) {

      synchronized (clients) {
        Iterator i = clients.iterator();
        while (  i.hasNext() ) {
          c = (DemeterClient)i.next();
          c.setTimer(c.getTimer()+SLEEP_TIME);
        System.out.println("Client: "+c.getTimer()+", against: "+(MOVIE_CREATION_TIME*60*60*60));
        if (c.getTimer() >= (MOVIE_CREATION_TIME * 60 *60 * 60)) {
            // Its time to do the movie
            try {
                make(c);
            } catch (Exception e) { }
            // When its done
            c.setTimer(0);
        }
      } }
      synchronized (waitThread) {
        try {
          waitThread.wait(SLEEP_TIME);
        } catch (InterruptedException e) { }
      }
    }
  }
}

/**
 * Helper thread. This thread handles the movie generation for the
 * given client.
 *
 */
class Worker   {

  private File root_dir;
  private File[] files;
  private Object waitThread = new Object();
  private boolean stop = false;
  private String  out;
  private Log log;

  /**
   * Starts the worker which assembles the movie.
   *
   * @param parent  The main movie generator.
   * @param image_dir The directory from which the movies are taken.
   * @param out The movie file.
   */
  public Worker(MovieCreator parent, String image_dir, String out)
     throws FileNotFoundException {

    this.log = parent.log;

    root_dir = new File(image_dir);

    if (!root_dir.exists())
      throw new FileNotFoundException ("Directory "+image_dir+" does not exist!");

      // We are only processing JPG movies.
    files = root_dir.listFiles(new JPGFilter());

    File movie_dir = new File(out);
    if (movie_dir.exists()) movie_dir.delete();

    movie_dir = new File(movie_dir.getParent());
    if (!movie_dir.isDirectory())
      throw new FileNotFoundException ("Directory "+
        movie_dir.getAbsolutePath()+" does not exist!");

    //movie_dir.mkdir();

    this.out = "file://"+out;

    this.run();
  }

  public void run() {

    Vector v = new Vector();
    JPGtoMOV conv;
    for (int i = 0; i< files.length; i++) {
      v.addElement(files[i].getPath());
    }

    // We must sort the movies in alphabetical order.
    java.util.Collections.sort((List)v);

    try {

      if (v.size()!=0) {
        // Open one of the movies and find out the widht+height

        FileInputStream fIn = new FileInputStream(files[0]);
        JPEGImageDecoder jpeg_decode = JPEGCodec.createJPEGDecoder(fIn);
        BufferedImage image = jpeg_decode.decodeAsBufferedImage();

        int width = image.getWidth();
        int height = image.getHeight();
        fIn.close();

        conv = new JPGtoMOV(log, out, width, height);
        conv.doIt(v);
      }
      // Erases the files.. after they have been processed.
     for (int i = 0; i< files.length; i++)
        files[i].delete();
    } catch (Exception e) {
      //e.printStackTrace();
      log.normalError(e.getMessage());
    }
  }
}

class JPGFilter implements FileFilter {

  public boolean accept (File f) {

    if (f.isDirectory()) {
            return true;
        }

        String extension = f.getName();
        int i = extension.lastIndexOf('.');

        if (i > 0 &&  i < extension.length() - 1) {
            extension = extension.substring(i+1).toLowerCase();
        }

        if (extension != null) {
            if (extension.equals("jpg") || extension.equals("jpeg")){
                    return true;
            } else {
                return false;
            }
        }

        return false;

  }
}
