package apollo.demeter;

/**
 * $Log: Liaison.java,v $
 * Revision 1.6  2002/04/28 19:14:23  uid500
 * Final updates.
 *
 * Revision 1.5  2002/04/16 19:22:32  uid500
 * Made it work with Hermes client count
 *
 * Revision 1.4  2002/04/14 01:25:41  uid500
 * Added comments.
 *
 * Revision 1.3  2002/04/07 03:37:27  uid500
 * Fixed the RTP start/close logic.
 *
 * Revision 1.2  2002/04/06 21:36:30  uid500
 * Fixed SecuritManger exception
 *
 * Revision 1.1  2002/04/02 02:11:38  uid500
 * Created the demeter daemons
 *
 * @author Konrad Rzeszutek (konrad@darnok.org)
 * @version $Id: Liaison.java,v 1.6 2002/04/28 19:14:23 uid500 Exp $
 */

import apollo.common.*;
import apollo.common.gui.*;
import apollo.common.client.*;
import apollo.common.movie.*;
import apollo.demeter.client.DemeterClient;
import apollo.hermes.rmi.*;
import apollo.common.*;
import apollo.ares.*;
import apollo.common.events.*;
import apollo.common.save.*;
import apollo.demeter.save.*;
import java.net.*;
import java.io.*;
import javax.media.*;

/**
 * Liaison - "One that maintains communication: served as the President's liaison with Congress."<br>
 *                                                        from www.dictionary.com<br><br>
 *
 * This object communicates with the Demeter RMI server. It receives
 * the orders to open (or close) a communication channel with a client.
 * Thus it setup's the connection, notifies the movie creator about a new
 * client, and adds the client to GUI.
 */
public class Liaison {

  protected String base_dir = null;
  protected ClientList clients;
  protected MainFrame f;
  protected MovieCreator m;
  protected Log log;
  protected ServerRMIInfo rmi;

  /**
   * Initializes the object.
   *
   * @param c The list of clients to be used when adding new clients
   * @param log The logging facility
   * @param g The GUI component. null if using -nogui flag.
   * @param m The movie generator.
   * @param base_dir The directory where images/movies will be stored/generated.
   */
  public Liaison(ClientList c, Log log,MainFrame g,MovieCreator m,
    String base_dir, ServerRMIInfo rmi) {

    this.base_dir = base_dir;
    this.clients = c;
    this.log = log;
    this.f = g;
    this.m = m;
    this.rmi = rmi;
  }

  /**
   * Register the client with the GUI and make the movie creator aware
   */
  public void registerClient(Client c) {

    clients.addElement(c);
    if (f != null) {
        f.addJInternalFrame(((DemeterClient)c).getFrame());
        if (f.clients != null)
          f.clients.refresh(clients);
    }
    m.register((DemeterClient)c);
  }

  /**
   * Start a new connection with the specified URL.
   */
  public void open(MediaLocator url) {

    // We use a thread so that the RMI server (which called this
    // method) doesn't have to wait.
    new WorkerOpen(this, url, base_dir).start();
  }

  /**
   * Close the connection with the specified URL.
   *
   * @throws IsNotRegisteredException if the URL was never used or activated.
   * @throws MalformedURLException if the given <code>url</code> is wrongly constructed.
   */
  public void close (MediaLocator url) throws IsNotRegisteredException,
  MalformedURLException {

    // This is silly, and about the only one, way to extract the hostname
    // from the URL. The MediaLocator.getURL() doesn't work - it throws an
    // MalformedURLExcpetion (when we have rtp:// in it).
    String host = (new URL("http:"+url.getRemainder()).getHost());
    DemeterClient client = null;

    // Find the server
    for (int i= 0; i< clients.length() ; i++) {
      if (clients.getElement(i).getHost().compareToIgnoreCase(host)==0) {
        client = (DemeterClient)clients.getElement(i);
        break;
      }
    }


    if (client == null)
      throw new IsNotRegisteredException("Wrong URL or host not "+
        "registered with the service.");
    // Start a thread to close neatly the client
    new WorkerClose(this, client).start();
  }

}

  /**
   * Helper threaded class to aid the Liaison in openning a new connection.
   *
   */
 class WorkerOpen extends Thread {

  private Liaison queen;
  private MediaLocator url;
  private String base_dir;

  /**
   *
   */
 public WorkerOpen(Liaison queen, MediaLocator url, String base_dir) {

    this.base_dir = base_dir;
    this.queen = queen;
    this.url = url;

  }

  /**
   * Starts the negotiation with the client. This is carried out
   * in the following order:<br>
   * <li>Create the RTPSender
   * <li>Create the DemeterClient
   * <li>Create DemeterFrame (which the DemeterClient actually does)
   * <li>Link both RTPSender and DemeterClient together
   * If everything went fine, <code>Liaison</code> is called back
   * notifying that the connection went fine and to finish the setup process.
   *
   */
  public void run () {

    try {

      String n = (new URL("http:"+url.getRemainder()).getHost());

      DemeterClient client = new DemeterClient(n,url.toExternalForm(),base_dir);

      createDirectory(client.getImageDir());
      createDirectory(client.getMovieDir());

      Codec[] codecs = new Codec[1];
      codecs[0] = new TimeStampEffect(client.getHost());
      Saver save = new RTPtoJPG(queen.log, client, codecs);
      save.open(url);

      client.registerSaver(save);
      save.registerClient((Client)client);
      queen.registerClient((Client)client);
      queen.rmi.setNrClients(queen.rmi.getNrClients()+1);

    } catch (Exception e) {
      e.printStackTrace();
      queen.log.println(e.getMessage(),Log.ERROR);
    }
  }
   private void createDirectory (String dir) {


        File[] dirs = null;
        File top = new File(dir);
        String step = top.getAbsolutePath();
        int cnt = 0;

        for (int i = 0; i< step.length(); i++)
          if (step.charAt(i) == '\\') cnt++;

        //System.out.println(step+","+cnt);
        dirs = new File[cnt+1];
        dirs[dirs.length-1] = top;
        for (int i = dirs.length-2; i>=0; i--)
          dirs[i] = new File(dirs[i+1].getParent());

        for (int i = 0; i< dirs.length; i++) {
          //System.out.println(dirs[i].getAbsolutePath());
          dirs[i].mkdir();
        }
  }
}
  /**
   * Helper thread used to close down the connection.
   */
  class WorkerClose extends Thread {

  private DemeterClient client;
  private Liaison queen;
  /**
   * Helper thread used to close down the connection
   *
   * @param queen The Liaison with all the required objects
   * @param clietn The client in question which has an existing
   *  Saver object that needs to be closed.
   */
  public WorkerClose(Liaison queen,DemeterClient client) {

    this.client = client;
    this.queen = queen;
  }

  /**
   *
   */
  public void run () {

    try {
      // Flush the RTP and close it
      client.getSaver().flush();
      client.getSaver().close();

      // Flush the movie production and close it
      queen.m.flush((DemeterClient)client);
      queen.m.remove((DemeterClient)client);

      // Remove the client from the GUI
      if (queen.f != null)
        queen.f.removeJInternalFrame(client.getFrame());

      queen.clients.removeElement((Client)client);

      // Close the client
      client.close();
      queen.rmi.setNrClients(queen.rmi.getNrClients()-1);
    } catch (Exception e) {
      queen.log.println(e.getMessage(),Log.ERROR);
    }
  }
}