package apollo.hermes;

/**
 * @author Konrad Rzeszutek (konrad@darnok.org)
 * @version $Id: HermesDaemon.java,v 1.5 2002/04/28 19:14:28 uid500 Exp $
 */
import apollo.common.*;
import apollo.common.gui.*;
import apollo.common.tui.*;
import apollo.common.jini.*;
import apollo.common.client.*;
import apollo.hermes.rmi.*;
import apollo.hermes.servers.*;
import apollo.hermes.gui.*;
import apollo.hermes.list.*;
import apollo.hermes.jini.JiniHermesInfo;
import apollo.demeter.jini.JiniDemeterInfo;
import apollo.mors.jini.JiniMorsInfo;
import apollo.nemesis.jini.JiniNemesisInfo;
import java.rmi.*;
import java.rmi.server.*;
import java.net.*;
import java.io.*;
import net.jini.lease.*;
import net.jini.core.lookup.*;
import net.jini.discovery.*;
import net.jini.core.discovery.*;
import net.jini.core.event.*;
import net.jini.core.lease.*;
import net.jini.core.lookup.ServiceItem;

/**
 * Hermes the messenger of Gods, guide of dead souls to Hades and
 * patron of travellers and thieves.
 * <br><br>
 * The purpose of this service is provide information (to any client)
 * about the capacity of servers, their address (RMI address), their load
 * and provide an easy way to find them.
 */
public class HermesDaemon implements DiscoveryListener, Daemon{

  private final static String DAEMON_NAME  = "HermesD";
  private final static String DAEMON_VER   = "$Id: HermesDaemon.java,v 1.5 2002/04/28 19:14:28 uid500 Exp $";

  /**
   * Maximum number of clients we can handle
   */
  public int DAEMON_MAX_CLIENTS = 32;

  protected Log log = new TextLog();

  /**
   * GUI components.
   */
  private MainFrame frame;
  private ServerListFrame[] l = null;

  protected ServersList servers;
  private String ver;
  private String hermes_url;
  private String RMI_server;
  private int RMI_port;
  private HermesRMIInfo hermes_rmi;

  private LookupDiscovery lookup              = null; //looks for all available lookup servers

  /**
   * Start the Hermes load-balancer.
   *
   * @param nogui If <code>true</code> service runs without the GUI.
   */
  public HermesDaemon(boolean nogui) {

  // Find out about ourselves - what's our IP, where the RMI server is,
    // on what port it resides
    getHostInfo();

    hermes_url = "rmi://"+RMI_server+":"+RMI_port+"/"+
      JiniServerComponentInfo.HERMES;


    if (!nogui) {
      frame = new MainFrame(DAEMON_NAME, this);
      frame.setSize(800,600);
      frame.validate();
      frame.setVisible(true);
      l = new ServerListFrame[3];
      l[0] = new ServerListFrame(JiniServerComponentInfo.MORS);
      l[1] = new ServerListFrame(JiniServerComponentInfo.NEMESIS);
      l[2] = new ServerListFrame(JiniServerComponentInfo.DEMETER);
      frame.addJInternalFrame(l[0]);
      frame.addJInternalFrame(l[1]);
      frame.addJInternalFrame(l[2]);
      Object waitFrame = new Object();
      while (frame.log == null) {
        synchronized (waitFrame) {
          try {
            waitFrame.wait(1000);
          } catch (InterruptedException e) { }
        }
      }
      log = ((Log)frame.log);
    }

    servers = new ServersList(l,log);

    // Jini:

    // Now add the Jini listener for the Jini services offered and call the RMI
    //this.leaser = new LeaseRenewalManager( );
    System.setSecurityManager(new RMISecurityManager( ));
    try{
      lookup = new LookupDiscovery(LookupDiscovery.ALL_GROUPS);
    }catch(java.io.IOException e){
        if (log.fatalError(e.getMessage())) System.exit(-1);
    }

    lookup.addDiscoveryListener(this);

    // Register our Jini for the clients
    JiniHermesInfo hermes_jini = new JiniHermesInfo(RMI_server, RMI_port);
    JiniServerRegister hermes_regi = new JiniServerRegister(
      new ServiceItem(null,hermes_jini,null),log);

    // Register our RMI services.

    try {
      hermes_rmi = new HermesRMIInfo();
    } catch (RemoteException e) {
      log.normalError(e.getMessage());
      //e.printStackTrace();
    }


    // Bind RMI
    try {
          log.println("Rebinding "+hermes_url,Log.INFO);
          Naming.rebind(hermes_url,hermes_rmi);
    } catch (Exception  e) {
      if (log.fatalError(e.getMessage())) exitDaemon();
      // Couldn't bind to RMI. Fatal error.
    }

    log.println("Started "+hermes_url+" server ..",Log.INFO);

    // Register our RMI server with the Server's list
    hermes_rmi.registerServersList(servers);

    log.println("Daemon started.",Log.INFO);
  }
  /**
   * Called when one or more lookup service registrars has been discovered.
   * The method should return quickly; e.g., it should not make remote calls.
   *
   * @param e - the event that describes the discovered registrars
   */
  public synchronized void discovered(DiscoveryEvent e){

    ServiceTemplate nemesis = new ServiceTemplate(null,
      new Class[]{JiniNemesisInfo.class}, null);
    ServiceTemplate mors = new ServiceTemplate(null,
      new Class[]{JiniMorsInfo.class}, null);
    ServiceTemplate demeter = new ServiceTemplate(null,
      new Class[]{JiniDemeterInfo.class}, null);

    ServiceRegistrar[ ] registrars = e.getRegistrars( );

    for(int i =0; i< registrars.length; i++){

      try{
          ServiceRegistrar registrar = registrars[i];
          log.println("Discovered "+registrar.getLocator().getHost()+
            " Jini server!","Jini");

          checkJiniFor(mors,    registrars[i]);
          checkJiniFor(demeter, registrars[i]);
          checkJiniFor(nemesis, registrars[i]);

	    }catch(Exception ex){
              log.normalError(ex.getMessage());
                //ex.printStackTrace();

	    }//end catch
	}//end outer for
    }//end method discovered

  /**
   * Checks the Jini for the <code>template</code> type server with the
   * given <code>registrar</code> and registers a JiniListner in case
   * the service appears at some point later on.
   *
   */
  private void checkJiniFor(ServiceTemplate template,ServiceRegistrar registrar)
    throws RemoteException{

        ServiceMatches services = registrar.lookup(template,Integer.MAX_VALUE);

        LookupLocator locator = registrar.getLocator( );

	/**
	 *  get the information from each one of the services
	 */
	for(int j =0; j< services.items.length; j++){
	  if(services.items[j].service != null){
                JiniServerComponentInfo info =
                  (JiniServerComponentInfo)services.items[j].service;
                Server s = new Server(info.getRMIName(),info.getRMIHost(),
                  info.getRMIPort());
                servers.addServer(s);
		}//end if
	    }//end for
          /**
	   * notify me of any other services that come in afterwards
	   */

          EventRegistration ereg = registrar.notify(template,
		ServiceRegistrar.TRANSITION_NOMATCH_MATCH,
		new JiniListner(template, servers), // Hmm for
		null,
                Lease.ANY);


    }
    /**
     * Called when one or more lookup service registrars has been discarded.
     *
     * @param e - the event that describes the discarded registrars
     */
    public synchronized void discarded(DiscoveryEvent e){
	ServiceRegistrar[] registrars = e.getRegistrars( );
	for(int i=0; i<registrars.length;i++){
	/**
	 * a condition where a Service Registrar
	 * leaves the vicinity.We cannot move the
	 * servers from the lookup server becuase they
	 * may actually be present on the other
	 * lookup servers.
	 */
 	    try{
 		ServiceRegistrar registrar = registrars[i];
		LookupLocator locator = registrar.getLocator( );
		log.println("The Lookup Service "+locator.getHost( )+
                  " has left.","Jini");
	    }catch(Exception ex){
		//System.out.println("Registrar could not leave");
                //ex.printStackTrace();
                log.println("Registrar could not leave","Jini");
	    }
	}//end for
    }//end method


 /**
   * Find out what our IP is. Where the RMI server resides, and what port
   * is the RMI server up on.
   */
  public void getHostInfo() {

      try {
        RMI_server = System.getProperty("RMISERVER",
          java.net.InetAddress.getLocalHost().getHostAddress());

      } catch (Exception e) {
        if (log.fatalError(e.getMessage())) exitDaemon();
      }
      String tmp = System.getProperty("RMIPORT","1098");
      RMI_port = 1098;
      try {
        RMI_port = Integer.parseInt(tmp);
      } catch (Exception e) { }
      ver = DAEMON_VER + "running under " + System.getProperty("os.name") +
        " " + System.getProperty("os.arch") +" " +System.getProperty("os.version");
  }

  /**
   * Safely terminates our daemon.
   *
   */
  public void exitDaemon() {

    servers.killThread();
    try {
      Naming.unbind(hermes_url);
    } catch (Exception e) { }
    System.exit(0);
  }


  public static void main(String[] args) {

    int i = 0;
    boolean nogui = false;

    while (i < args.length) {
      if (args[i].equals("-nogui")) {
		i++;
                nogui = true;
      } else if (args[i].equals("-h")) {
                System.out.println("java -jar hermes [-nogui]");
                System.exit(-2);
      }
      i++;
    }

    new HermesDaemon(nogui);

  }
}

