package apollo.hermes.list;

/**
 * @author Konrad Rzeszutek (konrad@darnok.org)
 * @version 1.0
 */

import apollo.hermes.rmi.*;
import apollo.hermes.servers.*;
import apollo.hermes.events.*;
import apollo.hermes.*;
import apollo.hermes.gui.ServerListFrame;
import apollo.common.Log;
import apollo.common.jini.JiniServerComponentInfo;

/**
 * Threaded list keeping track of the server's, also querying them (to
 * see if they are working and removing (if neccessary) the servers
 * which are unaccessible.
 */
public class ServersList extends Thread {

  /**
   * How often should the list of servers be checked. By default its
   * 9000 miliseconds (9 seconds).
   */
  public  int REFRESH_TIME = 9000;

  private ServerTypeList[] l;
  private boolean kill = false;
  private Object waitThread = new Object();
  private ServerListFrame[] frame;
  private Log log;

  /**
   * Initialize the list with the different type of server types.
   * This object is also threaded , so it starts off the threaded component.
   * <b>NOTE</b>: The list <code>f</code> is assumed to be of length three
   * and have initialized three ServerListFrame. Its assumed that <code>f</code>
   * list is in the order of MORS, NEMESIS and DEMETER.
   *
   * @param l  Three ServerListFrame - used to print out statistical information
   * @param log The logging component.
   *
   */
  public ServersList(ServerListFrame[] f, Log log) {

    l = new ServerTypeList[3];
    l[0] = new ServerTypeList(JiniServerComponentInfo.MORS);
    l[1] = new ServerTypeList(JiniServerComponentInfo.NEMESIS);
    l[2] = new ServerTypeList(JiniServerComponentInfo.DEMETER);

    frame = f;
    this.log = log;

    this.start();
  }


  /**
   * Adds a server to the list of Servers (sorted based on load) - and seperated by type
   *
   * @param  s	!= null
   */
  public void addServer(Server s) {

    log.println("New server: "+s.toString(), Log.INFO);
    for (int i = 0; i< l.length; i++)
      if (l[i].getType().compareToIgnoreCase(s.getType()) == 0) {
        l[i].addElement(s);
        if (frame != null)
          frame[i].refresh(l[i]);
        break;
      }

  }

  /**
   * Gets the server, based on <code>type</code> with the least load on it.
   *
   * @param type <code>Server.MORS</code> or <code>Server.NEMESIS</code> or <code>Server.DEMETER</code>
   * @return != null
   */
  public Server getServerWithLeastLoad(String type) {

  // ok. No searching here, just the first one from the list - the list is sorted by the thread.
    Server s = null;

    for (int i = 0; i< l.length; i++)
      if (l[i].getType().compareToIgnoreCase(type) == 0) {
        s = l[i].getWithLeastLoad();
        if (log != null && s != null) {
              log.println("Client asked for a server from "+type+
              " pool; found "+s.getHost()+" (Load: "+s.getLoad()+"%)",Log.INFO);
        }
        break;
      }
    return s;
  }

  /**
   * Receive a string representaion of the list.
   *
   * @return != null
   */
  public String toString() {

   String z = "";

   for (int i = 0; i< l.length; i++)
    z+= l[i].toString();

   return z;
  }



  /**
   * Safely terminate the thread.
   */
  public void killThread() {

    kill = true;

    synchronized (waitThread) {
      try {
        waitThread.notifyAll();
      } catch (Exception e) { }
    }
  }
  /**
   * Runs the thread. Internally it executes <code>refreshList</code> every @see #REFRESH_TIME.
   */
  public void run () {

    int j = 0;
    while (!kill) {
          try {

            for (j =0; j<l.length;j++)
                l[j].refreshList();
          } catch (ServerUnreachable e) {
            // This means one of the servers died unexpedtly.
            // Refrsh the list of servers.
            if (log != null) {
              log.println("Server "+e.getMessage()+" from the "+
                l[j].getType()+" pool is unreachable..",Log.INFO);
              //frame[j].refresh(l[j]);
            }
          }
          synchronized (waitThread) {
            try {
              waitThread.wait(REFRESH_TIME);
            } catch (Exception e) { }

          }
    }
  }
}
