package apollo.hermes.servers;

/**
 *
 * $Log: Server.java,v $
 * Revision 1.4  2002/04/14 01:26:42  uid500
 * Added comments
 *
 * Revision 1.3  2002/04/02 02:12:05  uid500
 * Modified
 *
 * Revision 1.2  2002/04/01 03:01:44  uid500
 * Took out stale code
 *
 * Revision 1.1.1.1  2002/03/30 18:51:32  uid500
 * Apollo src code import
 *
 * @author Konrad Rzeszutek (konrad@darnok.org)
 * @version $Id: Server.java,v 1.4 2002/04/14 01:26:42 uid500 Exp $
 */

import java.rmi.*;
import apollo.hermes.rmi.*;
import java.net.MalformedURLException;

/**
 * A server object (representing Demeter, Nemesis, or Mors) which
 * allows finding out the server's load, current capacity and
 * maximum amount of clients it can handle.
 *
 */
public class Server implements Comparable {


  protected String type = null;
  protected String hostname = null;
  protected int port = 0;
  protected int maxClients = -1;
  protected int nrClients = -1;
  private ServerRMIInfoInterface info;
  private boolean cache = false;
  /**
   * Initialize the server.
   *
   * @param h     The hostname of the server
   * @param p     The port on the server
   */
  public Server(String t,String h, int p) {
    type = t;
    hostname = h;
    port = p;
  }
  /**
   * Gets the unique String representation for the Server type.
   *
   * @return <code>MORS</code> || <code>NEMESIS</code> || <code>DEMETER</code>
   *        <code>NOT IMPLEMENTED</code>
   */
  public String getType() {
    if (type != null)
      return type;
    return "NOT IMPLEMENTED";
  }

  /**
   * Compares the <b>type</b> of the object, not the inner fields.
   *
   * @param   s   The compared object;
   * @return      <code>true</code> If the type is the same; Otherwise <code>
   *              false</code>.
   */
  public boolean equals(Server s) {

    return (s.type.equals(this.type));
  }

  /**
   * Gets the hostname of the <code>getType</code> server.
   *
   * @return    String != null
   */
  public String getHost() {

    return hostname;
  }
  /**
   * Gets the port of the <code>getType</code> server.
   *
   * @return    String != null
   */
  public int getPort() {

    return port;
  }

  /**
   * Gets the maximum amount of clients the <code>getType</code> server on
   * <code>getHost</code> system can support.
   *
   * @return    >=0
   */
  public  int getMaxClients() throws RemoteException, MalformedURLException, NotBoundException {

    if (hostname != null) {

      System.setSecurityManager(new RMISecurityManager());
      String url = "rmi://"+hostname+":"+port+"/"+type;

      if (!cache) info = null;
      cache = false;

      if (info == null)
        info = (ServerRMIInfoInterface)Naming.lookup(url);
      maxClients = info.getMaxClients();
      cache = true;

    }
    return maxClients;
  }
    /**
   * Gets the current number of clients the <code>getType</code> server on
   * <code>getHost</code> system is supporting right now.
   *
   * @return    >=0
   */
  public  int getNrClients() throws RemoteException, MalformedURLException, NotBoundException {
    if (hostname != null) {

      System.setSecurityManager(new RMISecurityManager());
      String url = "rmi://"+hostname+":"+port+"/"+type;


      if (!cache) info = null;
      cache = false;

      if (info == null)
        info = (ServerRMIInfoInterface)Naming.lookup(url);

      nrClients = info.getNrClients();
      cache = true;
    }
    return nrClients;
  }

  /**
   * Gets the load off the <code>getType</code> server on
   * <code>getHost</code> system.
   *
   * @return
   *      0 <= getLoad <= 100
   */
  public int getLoad()  {


    if (maxClients <= 0) // Don't want div by zero.
      return 100;

    return (int)(( (float)nrClients/(float)maxClients) * 100.0);

  }

  /**
   * Compare the load of this server to another.
   *
   * @param s - the specified Server to which we will compare
   * @return a negative integer, zero, or positive integer as this server has less,
   *  equal, or greater load than the specified object
   */
  public int compareTo(Object o) {

    int q = 0;
    int a = getLoad();
    int b = ((Server)o).getLoad();

    if (a < b) q = -1;
    if (a > b) q = 1;

    return q;

  }
  /**
   * Gets the String representation of this object.
   *
   * @return    != null
   */
  public String toString() {

    try {
      getMaxClients();
      getNrClients();

    } catch (Exception e) {
      nrClients = maxClients = -1;
    }

    return type+"@"+hostname+":"+port+" ("+nrClients+"/"+maxClients+") "+getLoad()+"% full";
  }
}
